# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import re
from tempfile import SpooledTemporaryFile
from typing import Any, Dict, Iterator, List, Tuple, Union, no_type_check
from urllib.parse import parse_qsl

from typing_extensions import Protocol

from .common import MultiDict
from .errors import FieldTooLarge, FileTooLarge, ParseError, TooManyFields
from .http import Headers, UploadedFile
from .typing import Header, RequestBody, RequestInput


class RequestParser(Protocol):  # pragma: no cover
    """Protocol for request parsers.
    """

    @property
    def mime_type(self) -> str:
        """Returns a string representing the mime type of the rendered
        content.  This is used to generate OpenAPI documents.
        """

    def can_parse_content(self, content_type: str) -> bool:
        """Returns True if this parser can parse the given content type.
        """

    @no_type_check
    def parse(self) -> Any:
        """Attempt to parse the input data.

        Raises:
          ParseError: if the data cannot be parsed.
        """


class JSONParser:
    """A JSON request parser.
    """

    mime_type = "application/json"

    def can_parse_content(self, content_type: str) -> bool:
        return content_type.startswith("application/json")

    def parse(self, data: RequestBody) -> Any:
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            raise ParseError("JSON input could not be parsed")


class URLEncodingParser:
    """A parser for urlencoded requests.
    """

    mime_type = "application/x-www-form-urlencoded"

    def can_parse_content(self, content_type: str) -> bool:
        return content_type.startswith("application/x-www-form-urlencoded")

    def parse(self, data: RequestBody) -> MultiDict[str, str]:
        try:
            return MultiDict(parse_qsl(data.decode("utf-8"), strict_parsing=True))
        except ValueError:
            raise ParseError("failed to parse urlencoded data")


class MultiPartParser:
    """A parser for multipart requests.  Returns a MultiDict mapping
    field names to lists of field string values or UploadedFiles.

    This is a reasonably simple streaming parser implementation for
    the multipart/form-data media type.  As such, it does not support
    deprecated parts of RFC7578 like multipart/mixed content and
    content-transfer-encoding headers.

    Parameters:
      bufsize: The max size of the streaming data buffer.  This should
        be a 32 bit integer that's a multiple of 4.  In some cases,
        the streaming data buffer may contain double this amount so
        take that into account when choosing a value.  Additionally,
        the value should be greater than the longest individual header
        value you want to accept.
      encoding: The codec to use when decoding form field values.
      encoding_errors: What to do when an decoding error is encountered.
      max_field_size: The max number of bytes a field can contain.
      max_file_size: The max number of bytes a file can contain.
      max_num_fields: The max number of fields accepted per request.
      max_spooled_size: The max number of bytes a file in the request
        can have before it's written to a temporary file on disk.
    """

    __slots__ = [
        "bufsize",
        "encoding",
        "encoding_errors",
        "max_field_size",
        "max_file_size",
        "max_num_fields",
        "max_spooled_size",
    ]

    BOUNDARY_RE = re.compile("boundary=(.+)")
    PARAMS_RE = re.compile('([A-Za-z]+)="([^"]+)"')

    mime_type = "multipart/form-data"

    def __init__(
            self, *,
            bufsize: int = 64 * 1024,
            encoding: str = "utf-8",
            encoding_errors: str = "replace",
            max_field_size: int = 500 * 1024,
            max_file_size: int = 10 * 1024 * 1024,
            max_num_fields: int = 100,
            max_spooled_size: int = 1024 * 1024,
    ) -> None:
        self.bufsize = bufsize
        self.encoding = encoding
        self.encoding_errors = encoding_errors
        self.max_field_size = max_field_size
        self.max_file_size = max_file_size
        self.max_num_fields = max_num_fields
        self.max_spooled_size = max_spooled_size

    def can_parse_content(self, content_type: str) -> bool:
        return content_type.startswith("multipart/form-data")

    def parse(self, content_type: Header, content_length: Header, body_file: RequestInput) -> MultiDict[str, Union[str, UploadedFile]]:  # noqa
        matches = self.BOUNDARY_RE.search(content_type)
        if not matches:
            raise ParseError("boundary missing from content-type header")

        boundary = matches.group(1)
        lines = self._iter_lines(body_file, boundary, int(content_length))
        parts = self._iter_parts(lines, boundary)
        return MultiDict(parts)

    def _iter_lines(self, stream: RequestInput, boundary: str, limit: int) -> Iterator[bytes]:
        buff = b""
        remaining = limit
        while remaining > 0:
            data = stream.read(self.bufsize)
            remaining -= len(data)
            if not data:
                return

            buff += data
            if remaining > 0 and len(buff) < self.bufsize:
                continue

            while buff:
                try:
                    i = buff.index(b"\r\n")
                except ValueError:
                    break

                line, buff = buff[:i + 2], buff[i + 2:]
                yield line

            if len(buff) >= self.bufsize and not buff.endswith(b"\r"):
                yield buff
                buff = b""

    def _iter_parts(self, lines: Iterator[bytes], boundary: str) -> Iterator[Tuple[str, Union[str, UploadedFile]]]:
        next_part = f"--{boundary}\r\n".encode()
        last_part = f"--{boundary}--\r\n".encode()

        def prepare_current_part() -> Tuple[str, Union[str, UploadedFile]]:
            nonlocal total_field_count
            headers = Headers(current_part_headers)
            name = current_part_disposition["name"]
            value: Union[str, UploadedFile]

            if "filename" in current_part_disposition:
                # Strip CRLF from the end of the file and then rewind.
                current_part_container.seek(-2, os.SEEK_END)
                current_part_container.truncate()

                headers.add("content-length", str(current_part_container.tell()))
                current_part_container.seek(0)

                filename = current_part_disposition["filename"]
                value = UploadedFile(filename, headers, current_part_container)
            else:
                # Strip CRLF from the end of the buffer.
                data = current_part_container[:-2]
                value = data.decode(self.encoding, errors=self.encoding_errors)

            total_field_count += 1
            return name, value

        def append_bytes(data: bytes) -> None:
            nonlocal current_part_container
            current_part_container += data

        total_field_count = 1
        current_part_bytes: int = 0
        current_part_is_file: bool = False
        current_part_container: Any = None
        current_part_writer: Any = None
        current_part_headers: Dict[str, Union[str, List[str]]] = {}
        current_part_disposition: Dict[str, str] = {}
        current_part_past_headers: bool = False
        for line in lines:
            if total_field_count > self.max_num_fields:
                raise TooManyFields("the input contains too many fields")

            if line == last_part:
                if current_part_container is not None:
                    yield prepare_current_part()

                break

            elif line == next_part:
                if current_part_container is not None:
                    yield prepare_current_part()

                current_part_bytes = 0
                current_part_is_file = False
                current_part_container = None
                current_part_writer = None
                current_part_headers = {}
                current_part_disposition = {}
                current_part_past_headers = False

            elif not current_part_past_headers:
                line = line.rstrip()
                if not line:
                    if current_part_container is None:
                        raise ParseError("content-disposition header is missing")

                    current_part_past_headers = True
                    continue

                header_name, _, header_value = line.decode().partition(": ")
                current_part_headers[header_name] = header_value

                if header_name.lower() == "content-disposition":
                    current_part_disposition = dict(self.PARAMS_RE.findall(header_value))

                    if "name" not in current_part_disposition:
                        raise ParseError("content-disposition header without a name")

                    if "filename" in current_part_disposition:
                        current_part_is_file = True
                        current_part_container = SpooledTemporaryFile(mode="wb+", max_size=self.max_spooled_size)
                        current_part_writer = current_part_container.write

                    else:
                        current_part_is_file = False
                        current_part_container = b""
                        current_part_writer = append_bytes

            else:
                current_part_bytes += len(line)
                if current_part_is_file and current_part_bytes >= self.max_file_size:
                    message = f"file '{current_part_disposition['name']}' exceeds the file size limit"
                    raise FileTooLarge(message)

                elif not current_part_is_file and current_part_bytes >= self.max_field_size:
                    message = f"field '{current_part_disposition['name']}' exceeds the field size limit"
                    raise FieldTooLarge(message)

                current_part_writer(line)

        else:
            raise ParseError("unexpected end of input")



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Here be Dragons.  OpenAPI and especially JSONSchema are not very
# good specs, but at least they're popular!

import ast
import inspect
from collections import defaultdict
from operator import itemgetter
from textwrap import dedent
from types import FunctionType
from typing import (
    Any, Callable, Dict, List, Optional, Set, Tuple, Union, get_type_hints, no_type_check
)

from typing_inspect import get_args, get_origin, is_generic_type, is_typevar

from ..app import BaseApp
from ..router import get_route_parameters
from ..typing import Header, QueryParam, extract_optional_annotation
from ..validation import Field, dump_schema, field, is_schema, schema


@schema
class Contact:
    """Contact information for the exposed API.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#contact-object
    """

    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


@schema
class License:
    """License information for the exposed API.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#license-object
    """

    name: str
    url: Optional[str] = None


@schema
class Metadata:
    """Metadata about the exposed API.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#infoObject
    """

    title: str
    description: str
    version: str

    terms_of_service: Optional[str] = field(response_name="termsOfService", default=None)
    contact: Optional[Contact] = None
    license: Optional[License] = None


@schema
class Schema:
    """Describes the type and attributes of a value.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#schemaObject
    """

    type: Optional[str] = None
    format: Optional[str] = None  # noqa
    description: Optional[str] = None
    required: Optional[List[str]] = None
    properties: Optional[Dict[str, Any]] = None
    nullable: Optional[bool] = None
    all_of: Optional[List[Dict[str, str]]] = field(response_name="allOf", default=None)
    any_of: Optional[List[Dict[str, str]]] = field(response_name="anyOf", default=None)
    items: Optional[Dict[str, str]] = None
    choices: Optional[List[str]] = field(response_name="enum", default=None)
    minimum: Optional[Union[int, float]] = None
    maximum: Optional[Union[int, float]] = None
    multiple_of: Optional[Union[int, float]] = field(response_name="multipleOf", default=None)
    min_length: Optional[int] = field(response_name="minLength", default=None)
    max_length: Optional[int] = field(response_name="maxLength", default=None)


@schema
class Parameter:
    """Describes a single handler parameter.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#parameterObject
    """

    name: str
    in_: str = field(response_name="in", choices=["query", "header", "path", "cookie"])
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: bool = False
    schema: Optional[Schema] = None


@schema
class APIKeySecurityScheme:
    """Describes an API key-based security scheme.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#securitySchemeObject
    """

    name: str
    in_: str = field(choices=["query", "header", "cookie"])
    type: str = field(response_only=True, default="apiKey")


@schema
class HTTPSecurityScheme:
    """Describes an HTTP-based security scheme.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#securitySchemeObject
    """

    name: str
    scheme: str = field(choices=["basic", "bearer"])
    type: str = field(response_only=True, default="http")


#: The union of acceptable security schemes.
SecurityScheme = Union[
    APIKeySecurityScheme,
    HTTPSecurityScheme,
]

@no_type_check  # noqa
def generate_openapi_document(
        app: BaseApp,
        metadata: Metadata,
        security_schemes: List[SecurityScheme],
        default_security_scheme: Optional[str] = None,
) -> Dict[str, Any]:
    """Generate an OpenAPI v3 document from an application object and
    some metadata.
    """
    request_mime_types = {parser.mime_type for parser in app.parsers}
    response_mime_types = {renderer.mime_type for renderer in app.renderers}

    schemas: Dict[str, Schema] = {}
    paths: Dict[str, Dict[str, Any]] = defaultdict(dict)
    for method, routes in app.router._routes_by_method.items():
        method = method.lower()

        for route in routes:
            handler = route.handler
            if not isinstance(handler, FunctionType):
                handler = handler.__call__  # type: ignore

            parameters = []
            request_schema_name = None
            annotations = get_type_hints(handler)
            route_template_parameters = get_route_parameters(route.template)
            for name, annotation in annotations.items():
                if name == "return":
                    continue

                is_optional, annotation = extract_optional_annotation(annotation)
                if name in route_template_parameters:
                    parameters.append(Parameter(
                        name, "path",
                        description=_get_annotation(handler, f"param_{name}_description"),
                        required=True,
                        schema=_generate_primitive_schema(annotation),
                    ))

                elif annotation is QueryParam:
                    parameters.append(Parameter(
                        name, "query",
                        description=_get_annotation(handler, f"param_{name}_description"),
                        required=not is_optional,
                        schema=Schema("string"),
                    ))

                elif annotation is Header:
                    parameters.append(Parameter(
                        name.replace("_", "-"), "header",
                        description=_get_annotation(handler, f"param_{name}_description"),
                        required=not is_optional,
                        schema=Schema("string"),
                    ))

                elif is_schema(annotation):
                    request_schema_name = _generate_schema("request", annotation, schemas)

            operation = paths[route.template][method] = {
                "tags": _get_annotation(handler, "tags", []),
                "operationId": route.name,
                "description": dedent(handler.__doc__ or "").rstrip(),
                "parameters": [dump_schema(param, sparse=True) for param in parameters],
                "deprecated": _get_annotation(handler, "deprecated", False),
                "responses": {"200": {"description": "A successful response.", "content": {}}},
            }

            if request_schema_name is not None:
                operation["requestBody"] = {"content": {}}
                for media_type in request_mime_types:
                    operation["requestBody"]["content"][media_type] = {
                        "schema": _make_schema_ref(request_schema_name),
                    }

                # Sort the media types for tools like the Swagger UI.
                operation["requestBody"]["content"] = _sort_dict(operation["requestBody"]["content"])

            # When the response annotation is a tuple of the form
            # [str, ...] then we assume that the responses will
            # contain custom status codes followed by the response
            # objects themselves.
            response_annotation = annotations.get("return")
            response_annotation_origin = get_origin(response_annotation)
            if response_annotation is not None and response_annotation_origin in (tuple, Tuple):
                arguments = get_args(response_annotation)
                if len(arguments) == 2 and arguments[0] is str and is_schema(arguments[1]):
                    response_annotation = arguments[1]

            if response_annotation is not None:
                if is_schema(response_annotation):
                    response_schema_name = _generate_schema("response", response_annotation, schemas)
                    for media_type in response_mime_types:
                        operation["responses"]["200"]["content"][media_type] = {
                            "schema": _make_schema_ref(response_schema_name),
                        }

                elif response_annotation_origin in (list, List):
                    arguments = get_args(response_annotation)
                    if is_schema(arguments[0]):
                        response_schema_name = _generate_schema("response", arguments[0], schemas)
                        for media_type in response_mime_types:
                            operation["responses"]["200"]["content"][media_type] = {
                                "schema": {
                                    "type": "array",
                                    "items": _make_schema_ref(response_schema_name),
                                }
                            }

            status_codes = _extract_status_codes(handler)
            for status_code in status_codes:
                status_code_ob = operation["responses"][str(status_code)] = {}
                if status_code < 300 and status_code != 204:
                    status_code_ob.update(**operation["responses"]["200"])

                description = _get_annotation(handler, f"response_{status_code}_description")
                status_code_ob.update(description=description)

            # If the declared return type is a response-code-tuple and
            # the status code finder couldn't find a 200 status code,
            # then it should be safe to drop that code from the
            # responses object.
            if get_origin(annotations.get("return")) in (tuple, Tuple) and 200 not in status_codes:
                del operation["responses"]["200"]

            # TODO: Add support for OAuth2 security scheme.
            # TODO: Make it possible to annotate that a handler
            # doesn't require a security scheme.
            if default_security_scheme is not None:
                operation["security"] = [{default_security_scheme: []}]

    return {
        "openapi": "3.0.1",
        "info": dump_schema(metadata, sparse=True),
        "paths": _sort_dict(paths),
        "components": {
            "schemas": _sort_dict({name: dump_schema(schema, sparse=True) for name, schema in schemas.items()}),
            "securitySchemes": _sort_dict({scheme.name: dump_schema(scheme, sparse=True) for scheme in security_schemes}),
        },
    }


@no_type_check
def _generate_schema(context: str, schema: Any, schemas: Dict[str, Schema]) -> str:
    name = f"{context}:{schema.__module__}.{schema.__name__}"
    if name in schemas:
        return name

    definition = Schema(
        "object",
        description=schema.__doc__,
        properties={},
        required=[],
    )

    for field in schema._FIELDS.values():  # noqa
        if context == "request":
            field_name = field.request_name

        elif context == "response":
            field_name = field.response_name

        else:  # pragma: no cover
            raise RuntimeError(f"invalid schema context {context!r}")

        # This is definitely covered but coverage.py seems to ignore
        # it for whatever reason.
        if context == "request" and field.response_only or \
           context == "response" and field.request_only:  # pragma: no cover
            continue

        is_optional, field_schema = _generate_field_schema(field, context, schemas)
        if field_schema is not None:
            definition.properties[field_name] = field_schema

        if not is_optional:
            definition.required.append(field_name)

    schemas[name] = definition
    return name


@no_type_check
def _generate_field_schema(field: Field, context: str, schemas: Dict[str, Schema]) -> Tuple[bool, Schema]:
    is_optional, annotation = extract_optional_annotation(field.annotation)
    if is_schema(annotation):
        field_schema_name = _generate_schema(context, annotation, schemas)
        field_schema = Schema(all_of=[_make_schema_ref(field_schema_name)])

    elif is_generic_type(annotation):
        origin = get_origin(annotation)
        if origin in (list, List):
            arguments = get_args(annotation)
            if not arguments or arguments == [Any] or is_typevar(arguments[0]):
                field_schema = _generate_primitive_schema(list)

            else:
                item_schema_name = _generate_schema(context, arguments[0], schemas)
                field_schema = Schema("array", items=_make_schema_ref(item_schema_name))

        elif origin in (dict, Dict):
            # TODO: Improve support for dicts.
            field_schema = _generate_primitive_schema(dict)

    else:
        field_schema = _generate_primitive_schema(annotation)

    if field_schema is not None:
        field_schema.description = field.description
        for option, value in field.validator_options.items():
            if option in Schema._FIELDS:
                setattr(field_schema, option, value)

    return is_optional, field_schema


@no_type_check
def _generate_primitive_schema(annotation: Any) -> Optional[Schema]:
    try:
        arguments = _PRIMITIVE_ANNOTATION_MAP[annotation]
        return Schema(*arguments)
    except KeyError:
        # TODO: Figure out a better way to handle this.
        return Schema("string")


def _extract_status_codes(handler: Callable[..., Any]) -> List[int]:
    try:
        source = inspect.getsource(handler)
        finder = _StatusCodeFinder()
        return finder.find(ast.parse(dedent(source)))
    except OSError:  # pragma: no cover
        return []


def _make_schema_ref(name: str) -> Dict[str, str]:
    return {"$ref": f"#/components/schemas/{name}"}


def _get_annotation(handler: Callable[..., Any], name: str, default: Any = None) -> Any:
    return getattr(handler, f"openapi_{name}", default)


def _sort_dict(data: Dict[Any, Any]) -> Dict[Any, Any]:
    # This relies on the ordered dict implementation in Py3.6+.
    return dict(sorted(data.items(), key=itemgetter(0)))


class _StatusCodeFinder(ast.NodeVisitor):
    """Finds usages of HTTP_* in an AST.
    """

    def __init__(self) -> None:
        self.status_codes: Set[int] = set()

    def find(self, tree: Any) -> List[int]:
        self.visit(tree)
        return sorted(list(self.status_codes))

    def visit_Name(self, node: Any) -> None:
        if node.id.startswith("HTTP_"):
            try:
                self.status_codes.add(int(node.id[len("HTTP_"):]))
            except ValueError:  # pragma: no cover
                pass


#: Maps primitive types to Schema parameters.
_PRIMITIVE_ANNOTATION_MAP = {
    int: ["integer", "int64"],
    str: ["string"],
    bool: ["boolean"],
    list: ["array"],
    dict: ["object"],
    float: ["number", "double"],
    bytes: ["string", "binary"],
}



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta
from typing import Dict, Optional, Union
from urllib.parse import parse_qsl, urlencode


class Cookies(Dict[str, str]):
    """A dictionary of request cookies.
    """

    @classmethod
    def parse(cls, cookie_header: str) -> "Cookies":
        """Turn a cookie header into a Cookies instance.
        """
        cookies = cls()
        cookie_strings = cookie_header.split(";")
        for cookie in cookie_strings:
            for name, value in parse_qsl(cookie.lstrip()):
                cookies[name] = value

        return cookies


class Cookie:
    """An individual response cookie.

    Raises:
      ValueError: If the value of same_site is not 'strict' or 'lax'.
    """

    __slots__ = [
        "name",
        "value",
        "max_age",
        "expires",
        "domain",
        "path",
        "secure",
        "http_only",
        "same_site",
    ]

    def __init__(
            self,
            name: str,
            value: str,
            max_age: Optional[Union[int, float, timedelta]] = None,
            expires: Optional[Union[int, float, datetime]] = None,
            domain: Optional[str] = None,
            path: Optional[str] = None,
            secure: bool = False,
            http_only: bool = False,
            same_site: Optional[str] = None,
    ) -> None:
        self.name = name
        self.value = value
        self.max_age = max_age
        self.expires = expires
        self.domain = domain
        self.path = path
        self.secure = secure
        self.http_only = http_only
        self.same_site = same_site

        if same_site and same_site != "strict" and same_site != "lax":
            raise ValueError("same_site must be either 'strict' or 'lax' or None")

    def encode(self) -> str:
        """Convert this cookie to a set-cookie header-compatible string.
        """
        output = [urlencode({self.name: self.value})]

        if self.max_age is not None:
            if isinstance(self.max_age, timedelta):
                duration = int(self.max_age.total_seconds())
            else:
                duration = int(self.max_age)

            output.append(f"Max-Age={duration}")

        if self.expires is not None:
            if isinstance(self.expires, (int, float)):
                expiration_date = datetime.utcfromtimestamp(self.expires)
            else:
                expiration_date = self.expires

            output.append(f"Expires={_format_cookie_date(expiration_date)}")

        if self.domain is not None:
            output.append(f"Domain={self.domain}")

        if self.path is not None:
            output.append(f"Path={self.path}")

        if self.secure:
            output.append("Secure")

        if self.http_only:
            output.append("HttpOnly")

        if self.same_site is not None:
            output.append(f"SameSite={self.same_site.title()}")

        return "; ".join(output)


_COOKIE_DATE_DAYS = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
_COOKIE_DATE_MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]


def _format_cookie_date(date: datetime) -> str:
    """Formats a cookie expiration date according to [1].

    [1]: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Date
    """
    tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, *_ = date.utctimetuple()
    day = _COOKIE_DATE_DAYS[tm_wday]
    month = _COOKIE_DATE_MONTHS[tm_mon - 1]
    return f"{day}, {tm_mday:02d}-{month}-{tm_year} {tm_hour:02d}:{tm_min:02d}:{tm_sec:02d} GMT"



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from io import BytesIO
from typing import Any

from molten import ParseError, RequestInput, Response, dump_schema, is_schema

try:
    import msgpack
except ImportError:  # pragma: no cover
    raise ImportError("'msgpack' package missing. Run 'pip install msgpack'.")


class MsgpackParser:
    """A msgpack_ request parser.

    .. _msgpack: https://msgpack.org/
    """

    mime_type = "application/x-msgpack"

    def can_parse_content(self, content_type: str) -> bool:
        return content_type.startswith("application/x-msgpack")

    def parse(self, body_file: RequestInput) -> Any:
        try:
            return next(msgpack.Unpacker(body_file, raw=False))
        except Exception as e:
            raise ParseError(f"msgpack input could not be parsed: {e}")


class MsgpackRenderer:
    """A msgpack_ response renderer.

    .. _msgpack: https://msgpack.org/
    """

    mime_type = "application/x-msgpack"

    def can_render_response(self, accept: str) -> bool:
        return accept.startswith("application/x-msgpack")

    def render(self, status: str, response_data: Any) -> Response:
        content = msgpack.packb(response_data, use_bin_type=True, default=self.default)
        return Response(status, stream=BytesIO(content), headers={
            "content-type": "application/x-msgpack",
        })

    def default(self, ob: Any) -> Any:
        """You may override this when subclassing the JSON renderer in
        order to encode non-standard object types.
        """
        if is_schema(type(ob)):
            return dump_schema(ob)

        raise TypeError(f"cannot encode values of type {type(ob)}")  # pragma: no cover



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import (
    Any, Callable, Dict, Generic, List, Optional, Sequence, Type, TypeVar, Union, no_type_check
)

from typing_extensions import Protocol
from typing_inspect import get_origin, is_generic_type, is_typevar

from ..errors import FieldValidationError, ValidationError
from ..typing import extract_optional_annotation
from .common import Missing, _Missing, is_schema

_T = TypeVar("_T")


@no_type_check
def field(*args, **kwargs) -> Any:
    """An alias for :class:`.Field` that tricks the type system into
    submission.
    """
    return Field(*args, **kwargs)


class Validator(Protocol[_T]):  # pragma: no cover
    """Validators ensure that values conform to arbitrary specifications.
    """

    def can_validate_field(self, field: "Field[_T]") -> bool:
        """This should return True if this validator can validate the given Field.
        """
        ...

    @no_type_check
    def validate(self, field: "Field[_T]", value: Any, **options: Any) -> _T:
        """Validate and possibly transform the given value.

        Raises:
          FieldValidationError: If the value is not valid.
        """
        ...


class Field(Generic[_T]):
    """An individual field on a schema.  The @schema decorator
    automatically turns annotated attributes into fields, but the
    field class can also be used to enrich annotated attributes with
    metadata.

    Examples:

      >>> @schema
      ... class Application:
      ...   name: str
      ...   rating: int = Field(minimum=1, maximum=5)

    Parameters:
      name: The name of the field.  Automatically populated by the
        schema decorator.
      annotation: The field's annotation.  Like name, this is
        automatically populated by @schema.
      description: An optional description for the field.
      default: An optional default value for the field.
      default_factory: An optional default function for the field.
      request_name: The field's name within a request.  This is the
        same as the field's name by default.
      response_name: The field's name within a response.  This is the
        same as the field's name by default.
      request_only: Whether or not to exclude this field from
        responses.  Defaults to False.
      response_only: Whether or not to ignore this field when loading
        requests.  Defaults to False.
      allow_coerce: Whether or not values passed to this field may be
        coerced to the correct type.  Defaults to False.
      validator: The validator to use when loading data.  The schema
        decorator will automatically pick a validator for builtin
        types.
      \**validator_options: Arbitrary options passed to the field's
        validator.
    """

    __slots__ = [
        "name",
        "annotation",
        "description",
        "default",
        "default_factory",
        "request_name",
        "response_name",
        "request_only",
        "response_only",
        "allow_coerce",
        "validator",
        "validator_options",
    ]

    def __init__(
            self,
            name: Optional[str] = None,
            annotation: Optional[Type[_T]] = None,
            description: Optional[str] = None,
            default: Union[_T, _Missing] = Missing,
            default_factory: Optional[Callable[[], _T]] = None,
            request_name: Optional[str] = None,
            response_name: Optional[str] = None,
            request_only: bool = False,
            response_only: bool = False,
            allow_coerce: bool = False,
            validator: Optional[Validator[_T]] = None,
            **validator_options: Any,
    ) -> None:
        self.name = name
        self.annotation = annotation
        self.description = description
        self.default = default
        self.default_factory = default_factory
        self.request_name = request_name or name
        self.response_name = response_name or name
        self.request_only = request_only
        self.response_only = response_only
        self.allow_coerce = allow_coerce
        self.validator = validator
        self.validator_options = validator_options

    def select_validator(self) -> None:
        """Find a suitable Validator for this field.
        """
        self.validator = _select_validator(self)
        if self.validator_options and not self.validator:
            raise RuntimeError(f"no validator could be selected for field {self}")

    @property
    def has_default(self) -> bool:
        """Returns True if the field has either a default value or a default factory.
        """
        return self.default is not Missing or self.default_factory is not None

    @no_type_check
    def validate(self, value: Optional[Any]) -> _T:
        """Validate and possibly transform the given value.

        Raises:
          FieldValidationError: When the value is not valid.
        """
        is_optional, annotation = extract_optional_annotation(self.annotation)
        # Distinguishing between missing values and null values is
        # important.  Optional types can have None as a value whereas
        # types with a default cannot.  Additionally, it's possible to
        # have an optional type without a default value.
        if value is Missing:
            if self.default is not Missing:
                return self.default

            elif self.default_factory:
                return self.default_factory()

            elif is_optional:
                return None

            raise FieldValidationError("this field is required")

        if value is None:
            if not is_optional:
                raise FieldValidationError("this field cannot be null")

            return value

        if not is_generic_type(annotation) and \
           not is_typevar(annotation) and \
           not is_schema(annotation) and \
           not isinstance(value, annotation):
            if not self.allow_coerce:
                raise FieldValidationError(f"unexpected type {type(value).__name__}")

            try:
                value = annotation(value)
            except Exception as e:
                raise FieldValidationError(f"value could not be coerced to {annotation.__name__}")

        if self.validator:
            return self.validator.validate(self, value, **self.validator_options)
        return value

    def __repr__(self) -> str:
        params = ", ".join(f"{name}={repr(getattr(self, name))}" for name in self.__slots__)
        return f"{type(self).__name__}({params})"


class NumberValidator:
    """Validates numbers.
    """

    def can_validate_field(self, field: Field[_T]) -> bool:
        _, annotation = extract_optional_annotation(field.annotation)
        return annotation is int or annotation is float

    def validate(
            self,
            field: Field[_T],
            value: Union[int, float],
            minimum: Optional[Union[int, float]] = None,
            maximum: Optional[Union[int, float]] = None,
            multiple_of: Optional[Union[int, float]] = None,
    ) -> Union[int, float]:
        if minimum is not None and value < minimum:
            raise FieldValidationError(f"value must be >= {minimum}")

        if maximum is not None and value > maximum:
            raise FieldValidationError(f"value must be <= {maximum}")

        if multiple_of is not None and value % multiple_of != 0:
            raise FieldValidationError(f"value must be a multiple of {multiple_of}")

        return value


class StringValidator:
    """Validates strings.
    """

    def can_validate_field(self, field: Field[_T]) -> bool:
        _, annotation = extract_optional_annotation(field.annotation)
        return annotation is str

    def validate(
            self,
            field: Field[_T],
            value: str,
            choices: Optional[Sequence[str]] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
    ) -> str:
        if choices is not None and value not in choices:
            raise FieldValidationError(f"must be one of: {', '.join(repr(choice) for choice in choices)}")

        if min_length is not None and len(value) < min_length:
            raise FieldValidationError(f"length must be >= {min_length}")

        if max_length is not None and len(value) > max_length:
            raise FieldValidationError(f"length must be <= {max_length}")

        return value


class ListValidator:
    """Validates lists.

    When a generic parameter is provided, then the values will be
    validated against that annotation::

      >>> @schema
      ... class Setting:
      ...   name: str
      ...   value: str

      >>> @schema
      ... class Account:
      ...   settings: List[Setting]

      >>> load_schema(Account, {"settings": [{"name": "a", "value": "b"}]})
      Account(settings=[Setting(name="a", value="b")])

      >>> load_schema(Account, {"settings": [{"name": "a"}]})
      Traceback (most recent call last):
        ...
      ValidationError: {"settings": {0: {"value": "this field is required"}}}

    When a generic parameter isn't provided, then any list is accepted.
    """

    def can_validate_field(self, field: Field[_T]) -> bool:
        _, annotation = extract_optional_annotation(field.annotation)
        return get_origin(annotation) in (list, List)

    @no_type_check
    def validate(
            self,
            field: Field[_T],
            value: List[Any],
            min_items: Optional[int] = None,
            max_items: Optional[int] = None,
            item_validator_options: Optional[Dict[str, Any]] = None,
    ) -> List[Any]:
        if not isinstance(value, list):
            raise FieldValidationError("value must be a list")

        if min_items is not None and len(value) < min_items:
            raise FieldValidationError(f"length must be >= {min_items}")

        if max_items is not None and len(value) > max_items:
            raise FieldValidationError(f"length must be <= {max_items}")

        # If the argument is Any, then the list can contain anything,
        # otherwise each item needs to be validated.
        annotation_args = getattr(field.annotation, "__args__", [])
        if annotation_args != (Any,):
            # This is a little piggy but it works well enough in practice.
            item_validator_options = item_validator_options or {}
            sub_field = Field(annotation=annotation_args[0], **item_validator_options)
            sub_field.select_validator()

            items = []
            for i, item in enumerate(value):
                try:
                    items.append(sub_field.validate(item))
                except FieldValidationError as e:
                    raise ValidationError({i: str(e)})
                except ValidationError as e:
                    raise ValidationError({i: e.reasons})

            return items

        return value


class DictValidator:
    """Validates dictionaries.

    When the ``fields`` option is provided, only the declared fields
    are going to be extracted from the input and will be validated.

      >>> @schema
      ... class Account:
      ...   settings: Dict[str, str] = Field(fields={
      ...     "a": Field(annotation=str),
      ...   })

      >>> load_schema(Account, {"settings": {}})
      Account(settings={})

      >>> load_schema(Account, {"settings": {"a": "b", "c": "d"}})
      Account(settings={"settings" {"a": "b"}})

      >>> load_schema(Account, {"settings": {"a": 42}})
      Traceback (most recent call last):
        ...
      ValidationError: {"settings": {"a": "unexpected type int"}}

    When the ``fields`` option is not provided and the annotation has
    generic parameters, then the items from the input will be
    validated against the generic parameter annotations::

      >>> @schema
      ... class Account:
      ...   settings: Dict[str, str]

      >>> load_schema(Account, {"settings": {}})
      Account(settings={})

      >>> load_schema(Account, {"settings": {"a": "b"}})
      Account(settings={"a": "b"})

      >>> load_schema(Account, {"settings": {"a": 42})  # invalid
      Traceback (most recent call last):
        ...
      ValidationError: {"settings": {"a": "unexpected type int"}}

    When neither ``fields`` or generic parameters are provided, then
    any dictionary will be accepted.
    """

    def can_validate_field(self, field: Field[_T]) -> bool:
        _, annotation = extract_optional_annotation(field.annotation)
        return get_origin(annotation) in (dict, Dict)

    @no_type_check
    def validate(
            self,
            field: Field[_T],
            value: Dict[Any, Any],
            fields: Optional[Dict[str, Field[Any]]] = None,
            key_validator_options: Optional[Dict[str, Any]] = None,
            value_validator_options: Optional[Dict[str, Any]] = None,
    ) -> Dict[Any, Any]:
        if not isinstance(value, dict):
            raise FieldValidationError("value must be a dict")

        # If a field dictionary was provided then we select specific
        # items from the input, otherwise we just validate the input.
        if fields is not None:
            items = {}
            for item_name, item_field in fields.items():
                try:
                    item_field.select_validator()
                    item_value = value.get(item_name, Missing)
                    items[item_name] = item_field.validate(item_value)
                except FieldValidationError as e:
                    raise ValidationError({item_name: str(e)})
                except ValidationError as e:
                    raise ValidationError({item_name: e.reasons})

            return items

        # If the args are [Any, Any], then the dict can contain
        # anything, otherwise each item needs to be validated.
        annotation_args = getattr(field.annotation, "__args__", [])
        if annotation_args and annotation_args != (Any, Any):
            key_validator_options = key_validator_options or {}
            key_field = Field(annotation=annotation_args[0], **key_validator_options)
            key_field.select_validator()

            value_validator_options = value_validator_options or {}
            value_field = Field(annotation=annotation_args[1], **value_validator_options)
            value_field.select_validator()

            items = {}
            for item_name, item_value in value.items():
                try:
                    item_name = key_field.validate(item_name)
                    item_value = value_field.validate(item_value)
                    items[item_name] = item_value
                except FieldValidationError as e:
                    raise ValidationError({item_name: str(e)})
                except ValidationError as e:
                    raise ValidationError({item_name: e.reasons})

            return items

        return value


class SchemaValidator:
    """Validates dictionaries against schema classes.
    """

    def can_validate_field(self, field: Field[_T]) -> bool:
        _, annotation = extract_optional_annotation(field.annotation)
        return is_schema(annotation)

    def validate(self, field: Field[_T], value: Dict[str, Any]) -> Any:
        from .schema import load_schema
        _, annotation = extract_optional_annotation(field.annotation)
        return load_schema(annotation, value)


#: The set of built-in validators.  Fields will attempt to use one of
#: these unless otherwise specified.
VALIDATORS: List[Validator[Any]] = [
    NumberValidator(),
    StringValidator(),
    ListValidator(),
    DictValidator(),
    SchemaValidator(),
]


def _select_validator(field: Field[_T]) -> Optional[Validator[_T]]:
    """Find a suitable validator for the given Field.
    """
    for validator in VALIDATORS:
        if validator.can_validate_field(field):
            return validator
    return None



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from collections import defaultdict
from typing import Any, Callable, Dict, Iterator, List, Optional, Pattern, Set, Tuple, Union

from .errors import RouteNotFound, RouteParamMissing

#: Alias for things that can be added to a router.
RouteLike = Union["Route", "Include"]


class Route:
    """An individual route.

    Examples:

      >>> Route("/accounts", list_accounts)
      >>> Route("/accounts", create_account, method="POST")
      >>> Route("/accounts/{account_id}", get_account)

    Parameters:
      template: A route template.
      handler: The request handler for this route.
      method: The request method.
      name: An optional name for the route.  Used in calls to
        reverse_uri.  Defaults to the name of the handler.
    """

    __slots__ = [
        "template",
        "handler",
        "method",
        "name",
    ]

    def __init__(self, template: str, handler: Callable[..., Any], method: str = "GET", name: Optional[str] = None) -> None:
        self.template = template
        self.handler = handler
        self.method = method
        self.name = name or handler.__name__


class Include:
    """Groups of routes prefixed by a common path.

    Examples:

      >>> Include("/v1/accounts", [
      ...   Route("/", create_account, method="POST"),
      ...   Route("/", list_accounts),
      ...   Route("/{account_id}", get_account),
      ... ])

    """

    __slots__ = [
        "prefix",
        "routes",
    ]

    def __init__(self, prefix: str, routes: List[RouteLike]) -> None:
        self.prefix = prefix
        self.routes = routes


class Router:
    """A collection of routes.
    """

    __slots__ = [
        "_routes_by_name",
        "_routes_by_method",
        "_route_res_by_method",
    ]

    def __init__(self, routes: Optional[List[RouteLike]] = None) -> None:
        self._routes_by_name: Dict[str, Route] = {}
        self._routes_by_method: Dict[str, List[Route]] = defaultdict(list)
        self._route_res_by_method: Dict[str, List[Pattern[str]]] = defaultdict(list)
        self.add_routes(routes or [])

    def add_route(self, route_like: RouteLike, prefix: str = "") -> None:
        """Add a Route to this instance.
        """
        if isinstance(route_like, Include):
            self.add_routes(route_like.routes, prefix + route_like.prefix)

        elif isinstance(route_like, Route):
            if route_like.name in self._routes_by_name:
                raise ValueError(f"a route named {route_like.name} is already registered")

            route = Route(
                template=prefix + route_like.template,
                handler=route_like.handler,
                method=route_like.method,
                name=route_like.name,
            )

            self._routes_by_name[route.name] = route
            self._routes_by_method[route.method].insert(0, route)
            self._route_res_by_method[route.method].insert(0, compile_route_template(route.template))

        else:  # pragma: no cover
            raise NotImplementedError(f"unhandled type {type(route_like)}")

    def add_routes(self, route_likes: List[RouteLike], prefix: str = "") -> None:
        """Add a set of routes to this instance.
        """
        for route_like in route_likes:
            self.add_route(route_like, prefix)

    def match(self, method: str, path: str) -> Union[None, Tuple[Route, Dict[str, str]]]:
        """Look up the route matching the given method and path.
        Returns the route and any path params that were extracted from
        the path.
        """
        routes = self._routes_by_method[method]
        route_res = self._route_res_by_method[method]
        for route, route_re in zip(routes, route_res):
            match = route_re.match(path)
            if match is not None:
                return route, match.groupdict()

        return None

    def reverse_uri(self, route_name: str, **params: str) -> str:
        """Build a URI from a Route.

        Raises:
          RouteNotFound: When the route doesn't exist.
          RouteParamMissing: When a required parameter was not provided.

        Parameters:
          route_name: The name of the route to reverse.
          \**params: Route params used to build up the path.
        """
        try:
            route = self._routes_by_name[route_name]
        except KeyError:
            raise RouteNotFound(route_name)

        uri = []
        for kind, token in tokenize_route_template(route.template):
            if kind == "binding" or kind == "glob":
                try:
                    uri.append(str(params[token]))
                except KeyError:
                    raise RouteParamMissing(token)

            elif kind == "chunk":
                uri.append(token)

        return "".join(uri)


def compile_route_template(template: str) -> Pattern[str]:
    """Convert a route template into a regular expression.
    """
    re_template = ""
    for kind, token in tokenize_route_template(template):
        if kind == "binding":
            re_template += f"(?P<{token}>[^/]+)"

        elif kind == "glob":
            re_template += f"(?P<{token}>.+)"

        elif kind == "chunk":
            re_template += token.replace(".", r"\.")

        else:  # pragma: no cover
            raise NotImplementedError(f"unhandled token kind {kind!r}")

    return re.compile(f"^{re_template}$")


def tokenize_route_template(template: str) -> Iterator[Tuple[str, str]]:
    """Convert a route template into a stream of tokens.
    """
    k, i = 0, 0

    while i < len(template):
        if template[i] == "{":
            yield "chunk", template[k:i]

            k = i
            kind = "binding"
            if template[i:i + 2] == "{*":
                kind = "glob"
                i += 1

            for j in range(i + 1, len(template)):
                if template[j] == "}":
                    yield kind, template[i + 1:j]
                    k = j + 1
                    i = j
                    break
            else:
                raise SyntaxError(f"unmatched {{ in route template {template!r}")

        i += 1

    if k != i:
        yield "chunk", template[k:i]


def get_route_parameters(template: str) -> Set[str]:
    """Extract all the named route parameters from a route template.
    """
    return {token for kind, token in tokenize_route_template(template) if kind == "binding"}



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict


class MoltenError(Exception):
    """Base class for all Molten exceptions.
    """

    __slots__ = ["status"]

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


class DIError(MoltenError):
    """Raised when a dependency cannot be resolved.
    """


class HTTPError(MoltenError):
    """Base class for HTTP errors.  Handlers and middleware can raise
    these to short-circuit execution.
    """

    __slots__ = ["status", "response"]

    def __init__(self, status: str, response: Any) -> None:
        self.status = status
        self.response = response

    def __str__(self) -> str:  # pragma: no cover
        return self.status


class RouteNotFound(MoltenError):
    """Raised when trying to reverse route to a route that doesn't exist.
    """


class RouteParamMissing(MoltenError):
    """Raised when a param is missing while reversing a route.
    """


class RequestParserNotAvailable(MoltenError):
    """Raised when no request parser can handle the incoming request.
    """


class ParseError(MoltenError):
    """Raised by parsers when the input data cannot be parsed.
    """


class FieldTooLarge(ParseError):
    """Raised by MultiPartParser when a field exceeds the maximum field size limit.
    """


class FileTooLarge(ParseError):
    """Raised by MultiPartParser when a file exceeds the maximum file size limit.
    """


class TooManyFields(ParseError):
    """Raised by MultiPartParser when the input contains too many fields.
    """


class HeaderMissing(MoltenError):
    """Raised by Headers.__getitem__ when a header does not exist.
    """


class ParamMissing(MoltenError):
    """Raised by QueryParams.__getitem__ when a param is missing.
    """


class ValidationError(MoltenError):
    """Raised by validator.load when the input data is invalid.
    """

    def __init__(self, reasons: Dict[str, Any]) -> None:
        self.reasons = reasons

    def __str__(self) -> str:  # pragma: no cover
        return str(self.reasons)


class FieldValidationError(MoltenError):
    """Raised by Field.validate when a given value is invalid.
    """



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
from typing import Any, Callable, Iterable, List, Optional
from wsgiref.util import FileWrapper  # type: ignore

from .components import (
    CookiesComponent, HeaderComponent, QueryParamComponent, RequestBodyComponent,
    RequestDataComponent, RouteParamsComponent, SchemaComponent
)
from .dependency_injection import Component, DependencyInjector
from .errors import RequestParserNotAvailable
from .http import HTTP_204, HTTP_404, HTTP_415, HTTP_500, Headers, QueryParams, Request, Response
from .middleware import ResponseRendererMiddleware
from .parsers import JSONParser, MultiPartParser, RequestParser, URLEncodingParser
from .renderers import JSONRenderer, ResponseRenderer
from .router import RouteLike, Router
from .typing import (
    Environ, Host, Method, Middleware, Port, QueryString, RequestInput, Scheme, StartResponse
)

LOGGER = logging.getLogger(__name__)


class BaseApp:
    """Base class for App implementations.

    Parameters:
      routes: An optional list of routes to register with the router.
      middleware: An optional list of middleware.  If provided, this
        replaces the default set of middleware, including the response
        renderer so make sure to include that in your middleware list.
      parsers: An optional list of request parsers to use.  If
        provided, this replaces the default list of request parsers.
      renderers: An optional list of response renderers.  If provided,
        this replaces the default list of response renderers.
    """

    def __init__(
            self,
            routes: Optional[List[RouteLike]] = None,
            middleware: Optional[List[Middleware]] = None,
            components: Optional[List[Component[Any]]] = None,
            parsers: Optional[List[RequestParser]] = None,
            renderers: Optional[List[ResponseRenderer]] = None,
    ) -> None:
        self.router = Router(routes)
        self.add_route = self.router.add_route
        self.add_routes = self.router.add_routes
        self.reverse_uri = self.router.reverse_uri

        self.parsers = parsers or [
            JSONParser(),
            URLEncodingParser(),
            MultiPartParser(),
        ]
        self.renderers = renderers or [
            JSONRenderer(),
        ]
        self.middleware = middleware or [
            ResponseRendererMiddleware(self.renderers)
        ]
        self.components = (components or []) + [
            HeaderComponent(),
            CookiesComponent(),
            QueryParamComponent(),
            RequestBodyComponent(),
            RequestDataComponent(self.parsers),
            SchemaComponent(),
        ]
        self.injector = DependencyInjector(self.components)

    def handle_404(self) -> Response:
        """Called whenever a route cannot be found.
        """
        return Response(HTTP_404, content="Not Found")

    def handle_415(self) -> Response:
        """Called whenever a request comes in with an unsupported
        content type.
        """
        return Response(HTTP_415, content="Unsupported Media Type")

    def handle_exception(self, exception: BaseException) -> Response:
        """Called whenever an unhandled exception occurs in middleware
        or a handler.  Dependencies are injected into this just like a
        normal handler.

        Parameters:
          exception: The exception that occurred.
        """
        LOGGER.exception("An unhandled exception occurred.")
        return Response(HTTP_500, content="Internal Server Error")

    def __call__(self, environ: Environ, start_response: StartResponse) -> Iterable[bytes]:  # pragma: no cover
        raise NotImplementedError("apps must implement '__call__'")


class App(BaseApp):
    """An application that implements the WSGI interface.

    Parameters:
      routes: An optional list of routes to register with the router.
      middleware: An optional list of middleware.  If provided, this
        replaces the default set of middleware, including the response
        renderer so make sure to include that in your middleware list.
      parsers: An optional list of request parsers to use.  If
        provided, this replaces the default list of request parsers.
      renderers: An optional list of response renderers.  If provided,
        this replaces the default list of response renderers.
    """

    def __call__(self, environ: Environ, start_response: StartResponse) -> Iterable[bytes]:
        request = Request.from_environ(environ)
        resolver = self.injector.get_resolver({
            BaseApp: self,
            Request: request,
            Method: Method(request.method),
            Scheme: Scheme(request.scheme),
            Host: Host(request.host),
            Port: Port(request.port),
            QueryString: QueryString(environ.get("QUERY_STRING", "")),
            QueryParams: request.params,
            Headers: request.headers,
            RequestInput: RequestInput(request.body_file),
        })

        try:
            handler: Callable[..., Any]
            route_and_params = self.router.match(request.method, request.path)
            if route_and_params is not None:
                route, params = route_and_params
                handler = route.handler
                resolver.add_component(RouteParamsComponent(params))
            else:
                handler = self.handle_404

            handler = resolver.resolve(handler)
            for middleware in reversed(self.middleware):
                handler = resolver.resolve(middleware(handler))

            exc_info = None
            response = handler()
        except RequestParserNotAvailable:
            exc_info = None
            response = resolver.resolve(self.handle_415)()
        except Exception as e:
            exc_info = sys.exc_info()
            response = resolver.resolve(self.handle_exception)(exception=e)

        content_length = response.get_content_length()
        if content_length is not None:
            response.headers.add("content-length", str(content_length))

        start_response(response.status, list(response.headers), exc_info)
        if response.status != HTTP_204:
            wrapper = environ.get("wsgi.file_wrapper", FileWrapper)
            return wrapper(response.stream)
        else:
            return []



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from inspect import Parameter
from typing import Optional

from molten import Settings as Settings

try:
    import toml
except ImportError:  # pragma: no cover
    raise ImportError("'toml' package missing. Run 'pip install toml'.")


#: Canary value representing missing values.
Missing = object()


class TOMLSettings(Settings):
    """A dictionary of settings parsed from a TOML file.
    """

    @classmethod
    def from_path(cls, path: str, environment: str) -> "Settings":
        """Load a TOML file into a dictionary.

        Raises:
          FileNotFoundError: When the settings file does not exist.

        Parameters:
          path: The path to the TOML file containing your settings.
          environment: The config environment to use.
        """
        all_settings = toml.load(open(path))
        common_settings = all_settings.get("common", {})
        environment_settings = all_settings.get(environment, {})
        return cls({**common_settings, **environment_settings})


class TOMLSettingsComponent:
    """A component that loads settings from a TOML file.

    The settings file should have a "common" section and one section
    for each environment the application is expected to run in.  The
    environment-specific settings are merged on top of the "common"
    settings on load.

    The current environment is determined by the ``ENVIRONMENT``
    config variable.

    Example settings file::

      [common]
      conn_pooling = true
      conn_pool_size = 1

      [dev]

      [prod]
      con_pool_size = 32

    Examples::

      from molten import Settings

      def handle(settings: Settings):
        settings.get("conn_pooling")

    Parameters:
      path: The path to the TOML file containing your settings.
      environment: The config environment to use.  If not provided,
        this defaults to the value of the "ENVIRONMENT" environment
        variable.  If that's not set either, then this defaults to
        "dev".
    """

    __slots__ = ["path", "environment"]

    is_cacheable = True
    is_singleton = True

    def __init__(self, path: str = "./settings.toml", environment: Optional[str] = None) -> None:
        self.path = path
        self.environment = environment or os.getenv("ENVIRONMENT", "dev")

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Settings

    def resolve(self) -> Settings:
        return TOMLSettings.from_path(self.path, self.environment)



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union
from urllib.parse import urlencode

from ..http import Headers, QueryParams, Request
from ..typing import Environ


def to_environ(ob: Union[Request, Headers, QueryParams]) -> Environ:
    """Convert request abstractions to WSGI environ dicts.
    """

    if isinstance(ob, Request):
        return {
            "HTTP_HOST": ob.host,
            "PATH_INFO": ob.path,
            "REQUEST_METHOD": ob.method,
            "SERVER_PORT": ob.port,
            "wsgi.input": ob.body_file,
            "wsgi.url_scheme": ob.scheme,
            **to_environ(ob.params),
            **to_environ(ob.headers),
        }

    elif isinstance(ob, Headers):
        headers = {f"HTTP_{name.upper().replace('-', '_')}": value for name, value in ob}

        try:
            headers["CONTENT_TYPE"] = headers.pop("HTTP_CONTENT_TYPE")
        except KeyError:
            pass

        try:
            headers["CONTENT_LENGTH"] = headers.pop("HTTP_CONTENT_LENGTH")
        except KeyError:
            pass

        return headers

    elif isinstance(ob, QueryParams):
        return {"QUERY_STRING": urlencode(list(ob))}

    else:  # pragma: no cover
        raise NotImplementedError(f"to_environ cannot handle type {type(ob)}")



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional, Type, TypeVar, get_type_hints, no_type_check

from ..errors import FieldValidationError, ValidationError
from .common import Missing, is_schema
from .field import Field

_T = TypeVar("_T")


def schema(cls: Type[_T]) -> Type[_T]:
    """Construct a schema from a class.

    Schemas are plain Python classes with automatically-generated
    ``__init__``, ``__eq__`` and ``__repr__`` methods.  They may be
    used to validate requests and serialize responses.

    Examples:

      >>> @schema
      ... class Account:
      ...   username: str
      ...   password: str = Field(request_only=True)
      ...   is_admin: bool = Field(response_only=True, default=False)

      >>> load_schema(Account, {})
      Traceback (most recent call last):
        ...
      ValidationError: {'username': 'this field is required', 'password': 'this field is required'}

      >>> load_schema(Account, {"username": "example", "password": "secret"})
      Account(username='example', password='secret', is_admin=False)

      >>> dump_schema(load_schema(Account, {"username": "example", "password": "secret"}))
      {'username': 'example', 'is_admin': False}

    Raises:
      RuntimeError: When the attributes are invalid.
    """
    fields = {}
    for base in cls.__mro__[-1:0:-1]:
        base_fields = getattr(base, "_FIELDS", {})
        for name, field in base_fields.items():
            fields[name] = field

    annotations = get_type_hints(cls)
    found_default = False
    for name, annotation in annotations.items():
        value = getattr(cls, name, Missing)
        if value is Missing:
            value = fields.get(name, value)

        if isinstance(value, Field):
            value.name = name
            value.annotation = annotation
            value.request_name = value.request_name or name
            value.response_name = value.response_name or name

            fields[name] = value

        else:
            fields[name] = Field(name=name, annotation=annotation, default=value)

        # At this point the field instance has an annotation for sure
        # so it's safe to select a Validator.
        field = fields[name]
        field.select_validator()

        # Make sure fields without a default don't come after fields
        # with one.
        if field.has_default:
            found_default = True

        elif found_default:
            raise RuntimeError("attributes without a default cannot follow ones with a default")

        # Remove the attribute from the class definition.
        try:
            if value is not Missing:
                delattr(cls, name)
        except AttributeError:
            pass

    if not fields:
        raise RuntimeError(f"schema {cls.__name__} doesn't have any fields")

    setattr(cls, "__slots__", list(fields))
    setattr(cls, "_SCHEMA", True)
    setattr(cls, "_FIELDS", fields)
    _add_init(cls, fields)
    _add_fn(cls, "__eq__", ["self", "other"], _EQ_FN_BODY)
    _add_fn(cls, "__repr__", ["self"], _REPR_FN_BODY)
    return cls


@no_type_check
def load_schema(schema: Type[_T], data: Dict[str, Any]) -> _T:
    """Validate the given data dictionary against a schema and
    instantiate the schema.

    Raises:
      ValidationError: When the input data is not valid.

    Parameters:
      schema: The schema class to validate the data against.
      data: Data to validate against and populate the schema with.
    """
    if not is_schema(schema):
        raise TypeError(f"{schema} is not a schema")

    errors, params = {}, {}
    for field in schema._FIELDS.values():
        if field.response_only:
            # Response-only fields without an explicit default have to
            # default to _something_ so we choose None.
            if not field.has_default:
                params[field.name] = None

            continue

        try:
            value = data.get(field.request_name, Missing)
            params[field.name] = field.validate(value)
        except FieldValidationError as e:
            errors[field.request_name] = str(e)
        except ValidationError as e:
            errors[field.request_name] = e.reasons

    if errors:
        raise ValidationError(errors)

    return schema(**params)


@no_type_check
def dump_schema(ob: Any, *, sparse: bool = False) -> Dict[str, Any]:
    """Convert a schema instance into a dictionary.

    Raises:
      TypeError: If ob is not a schema instance.

    Parameters:
      ob: An instance of a schema.
      sparse: If true, fields whose values are None are going to be
        dropped from the output.
    """
    if not is_schema(type(ob)):
        raise TypeError(f"{ob} is not a schema")

    data = {}
    for field in ob._FIELDS.values():
        if field.request_only:
            continue

        value = getattr(ob, field.name)
        if is_schema(type(value)):
            value = dump_schema(value, sparse=sparse)

        elif isinstance(value, list):
            value = [dump_schema(item, sparse=sparse) if is_schema(type(item)) else item for item in value]

        elif isinstance(value, dict):
            value = {name: dump_schema(item, sparse=sparse) if is_schema(type(item)) else item for name, item in value.items()}

        if sparse and value is None:
            continue

        data[field.response_name] = value

    return data


def _add_fn(
        cls: Type[Any],
        name: str,
        params: List[str],
        body: List[str],
        fn_globals: Optional[Dict[str, Any]] = None,
        fn_locals: Optional[Dict[str, Any]] = None,
) -> None:
    """Construct a function and add it to a class.
    """
    if name in cls.__dict__:
        return

    fn_globals = {"Missing": Missing, **(fn_globals or {})}
    fn_locals = fn_locals or {}
    definition = _FN_TEMPLATE.format(
        name=name,
        params=", ".join(params),
        body="\n    ".join(body),
    )

    exec(definition, fn_globals, fn_locals)
    setattr(cls, name, fn_locals[name])


def _add_init(cls: Type[Any], fields: Dict[str, Field[_T]]) -> None:
    """Construct and add an init function to a schema.
    """
    fn_globals: Dict[str, Any] = {}
    fn_params = ["self"]
    fn_body = []
    for field in fields.values():
        if field.default is not Missing:
            default_name = f"_{field.name}_default"
            fn_globals[default_name] = field.default
            fn_params.append(f"{field.name}=Missing")
            fn_body.append(f"self.{field.name} = {field.name} if {field.name} is not Missing else {default_name}")

        elif field.default_factory:
            factory_name = f"_{field.name}_default_factory"
            fn_globals[factory_name] = field.default_factory
            fn_params.append(f"{field.name}=Missing")
            fn_body.append(f"self.{field.name} = {field.name} if {field.name} is not Missing else {factory_name}()")

        else:
            fn_params.append(f"{field.name}")
            fn_body.append(f"self.{field.name} = {field.name}")

    _add_fn(cls, "__init__", fn_params, fn_body, fn_globals)


_FN_TEMPLATE = """\
def {name}({params}):
    {body}
""".rstrip()

_EQ_FN_BODY = """\
try:
    return all(getattr(self, name) == getattr(other, name) for name in self._FIELDS)
except AttributeError:
    return False
""".rstrip().split("\n")

_REPR_FN_BODY = """\
params = ', '.join(f'{name}={repr(getattr(self, name))}' for name in self._FIELDS)
return f'{type(self).__name__}({params})'
""".rstrip().split("\n")



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['molten',
 'molten.http',
 'molten.openapi',
 'molten.openapi.templates',
 'molten.testing',
 'molten.validation']

package_data = \
{'': ['*'], 'molten': ['contrib/*']}

install_requires = \
['typing-extensions>=3.6,<4.0', 'typing-inspect>=0.3.1,<0.4.0']

setup_kwargs = {
    'name': 'molten',
    'version': '0.2.0',
    'description': 'An experimental web framework for Python 3.6+.',
    'long_description': "# molten\n\n[![Build Status](https://travis-ci.org/Bogdanp/molten.svg?branch=master)](https://travis-ci.org/Bogdanp/molten)\n[![Maintainability](https://api.codeclimate.com/v1/badges/f7ab5fdd188fda33c3e1/maintainability)](https://codeclimate.com/github/Bogdanp/molten/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/f7ab5fdd188fda33c3e1/test_coverage)](https://codeclimate.com/github/Bogdanp/molten/test_coverage)\n[![PyPI version](https://badge.fury.io/py/molten.svg)](https://badge.fury.io/py/molten)\n[![Documentation](https://img.shields.io/badge/doc-latest-brightgreen.svg)](https://moltenframework.com)\n\n**molten** is an experimental web framework for Python 3.6+.  It's\nmeant to be minimal, extensible, fast and productive.  Check out the\n[examples] for a taste or start reading the [user guide]!\n\n\n## Installation\n\n    pip install molten\n\n\n## License\n\nmolten is licensed under the LGPL.  Please see [COPYING] and\n[COPYING.LESSER] for licensing details.\n\n[COPYING.LESSER]: https://github.com/Bogdanp/molten/blob/master/COPYING.LESSER\n[COPYING]: https://github.com/Bogdanp/molten/blob/master/COPYING\n[examples]: https://github.com/Bogdanp/molten/blob/master/examples\n[user guide]: https://moltenframework.com/guide.html\n",
    'author': 'Bogdan Popa',
    'author_email': 'bogdan@cleartype.io',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
